"""Management of versions to help users update."""

import os
from configparser import ConfigParser
from datetime import datetime

from distutils.version import StrictVersion
from .utils.logging import print_warning  # pylint: disable=cyclic-import


def get_latest_version() -> str:
    """Get latest version from PyPI."""
    # pylint: disable=import-outside-toplevel
    import requests  # type: ignore

    url = "https://pypi.org/pypi/redbrick-sdk/json"
    data = requests.get(url).json()
    versions = list(data["releases"].keys())
    versions.sort(key=StrictVersion)
    return str(versions[-1])


def version_check(current_version: str) -> None:
    """Check if current installed version of the SDK is up to date with latest pypi release."""
    cache_file = os.path.join(os.path.expanduser("~"), ".redbrickai", "version")
    os.makedirs(os.path.dirname(cache_file), exist_ok=True)

    cache_config = ConfigParser()
    cache_config.read(cache_file)

    update_cache = False

    if (
        "version" not in cache_config
        or "current_version" not in cache_config["version"]
        or cache_config["version"]["current_version"] != current_version
    ):
        cache_config["version"] = {"current_version": current_version}
        update_cache = True

    current_timestamp = int(datetime.now().timestamp())

    if (
        "latest_version" not in cache_config["version"]
        or "last_checked" not in cache_config["version"]
        or current_timestamp - int(cache_config["version"]["last_checked"]) > 86400
    ):
        latest_version = get_latest_version()
        # Comparing with current installed version
        if current_version != latest_version:
            warn = (
                "You are using version '{}' of the SDK. However, version '{}' is available!\n"
                + "Please update as soon as possible to get the latest features and bug fixes.\n"
                + "You can use 'python -m pip install --upgrade redbrick-sdk'"
                + " to get the latest version."
            )
            print_warning(warn.format(current_version, latest_version))

        cache_config["version"]["latest_version"] = latest_version
        cache_config["version"]["last_checked"] = str(current_timestamp)
        update_cache = True

    if update_cache:
        with open(cache_file, "w", encoding="utf-8") as file_:
            cache_config.write(file_)



"""Common utility functions."""
import hashlib
from typing import Union


def hash_sha256(message: Union[str, bytes]) -> str:
    """Return basic SHA256 of given message."""
    sha256 = hashlib.sha256()
    sha256.update(message.encode() if isinstance(message, str) else message)
    return sha256.hexdigest()



"""Main file for CLI."""
import sys
import argparse
from typing import Any, List, Optional

import redbrick
from redbrick.cli.command import (
    CLIConfigController,
    CLIInitController,
    CLICloneController,
    CLIInfoController,
    CLIExportController,
    CLIPruneController,
    CLIUploadController,
)
from redbrick.utils.logging import print_warning, handle_exception

from .cli_base import CLIInterface


class CLIController(CLIInterface):
    """Main CLI Controller."""

    def __init__(self, command: argparse._SubParsersAction) -> None:
        """Initialize CLI command parsers."""
        self.config = CLIConfigController(
            command.add_parser(self.CONFIG, help="Setup credentials")
        )
        self.init = CLIInitController(
            command.add_parser(self.INIT, help="Create a new project")
        )
        self.clone = CLICloneController(
            command.add_parser(
                self.CLONE, help="Clone an existing remote project to local"
            )
        )
        self.info = CLIInfoController(
            command.add_parser(self.INFO, help="Get a project's information")
        )
        self.export = CLIExportController(
            command.add_parser(self.EXPORT, help="Export data for a project")
        )
        self.prune = CLIPruneController(
            command.add_parser(self.PRUNE, help="Remove old export files")
        )
        self.upload = CLIUploadController(
            command.add_parser(self.UPLOAD, help="Upload files to a project")
        )

    def handle_command(self, args: argparse.Namespace) -> None:
        """CLI command main handler."""
        if args.command == self.CONFIG:
            self.config.handler(args)
        elif args.command == self.INIT:
            self.init.handler(args)
        elif args.command == self.CLONE:
            self.clone.handler(args)
        elif args.command == self.INFO:
            self.info.handler(args)
        elif args.command == self.EXPORT:
            self.export.handler(args)
        elif args.command == self.PRUNE:
            self.prune.handler(args)
        elif args.command == self.UPLOAD:
            self.upload.handler(args)
        else:
            raise argparse.ArgumentError(None, "")


def cli_parser(generate_docs: bool = True) -> Any:
    """Initialize argument parser."""
    parser = argparse.ArgumentParser(description="RedBrick AI")
    parser.add_argument(
        "-v", "--version", action="version", version=f"v{redbrick.__version__}"
    )
    cli = CLIController(parser.add_subparsers(title="Commands", dest="command"))

    if generate_docs:
        return parser

    return parser, cli


@handle_exception
def cli_main(argv: Optional[List[str]] = None) -> None:
    """CLI main handler."""
    parser: argparse.ArgumentParser
    cli: CLIController

    parser, cli = cli_parser(False)

    try:
        args = parser.parse_args(argv if argv is not None else sys.argv[1:])
    except argparse.ArgumentError as error:
        print_warning(str(error))
        parser.print_help()
    else:
        try:
            cli.handle_command(args)
        except argparse.ArgumentError:
            parser.print_usage()



"""Utilities for working with label objects."""

from typing import Dict, List, Optional


def clean_rb_label(label: Dict) -> Dict:
    """Clean any None fields."""
    for key, val in label.copy().items():
        if val is None:
            del label[key]
    return label


def flat_rb_format(
    labels: List[Dict],
    items: List[str],
    items_presigned: List[str],
    name: str,
    created_by: str,
    task_id: str,
    current_stage_name: str,
    labels_map: Optional[List[Dict]],
) -> Dict:
    """Get standard rb flat format, same as import format."""
    return {
        "labels": labels,
        "items": items,
        "itemsPresigned": items_presigned,
        "name": name,
        "taskId": task_id,
        "createdBy": created_by,
        "currentStageName": current_stage_name,
        "labelsMap": labels_map or [],
    }


def dicom_rb_format(task: Dict) -> Dict:
    """Get new dicom rb task format."""
    # pylint: disable=too-many-branches
    output = {"taskId": task["taskId"]}

    if task.get("name"):
        output["name"] = task["name"]
    elif task.get("items"):
        output["name"] = task["items"][0]

    if task.get("createdBy"):
        output["createdBy"] = task["createdBy"]

    if task.get("currentStageName"):
        output["currentStageName"] = task["currentStageName"]

    if task.get("items"):
        output["items"] = task["items"]

    if task.get("itemsPresigned"):
        output["itemsPresigned"] = task["itemsPresigned"]

    if task.get("labelsMap"):
        input_labels_map = task["labelsMap"]
        segmentations = {
            str(label_map["imageIndex"]): label_map["labelName"]
            for label_map in input_labels_map
            if label_map["labelName"]
        }
        keys = list(map(int, segmentations.keys()))

        if sorted(keys) == list(range(len(input_labels_map))):
            output["segmentations"] = [None for _ in keys]
            for label_map in input_labels_map:
                output["segmentations"][label_map["imageIndex"]] = label_map[
                    "labelName"
                ]
        elif keys:
            output["segmentations"] = segmentations

    if task.get("labels"):
        input_labels = task["labels"]
        segment_map = {}
        labels = []
        for label in input_labels:
            if label.get("dicom", {}).get("instanceid"):
                segment_map[str(label["dicom"]["instanceid"])] = (
                    label["category"][0][1]
                    if len(label["category"][0]) == 2
                    else label["category"][0][1:]
                )
            else:
                labels.append(clean_rb_label(label))

        if segment_map:
            output["segmentMap"] = segment_map

        if labels:
            output["labels"] = labels

    return output



"""Graphql Client responsible for make API requests."""
from typing import Dict
import requests  # type: ignore

import aiohttp
import tenacity

from redbrick import __version__ as sdk_version  # pylint: disable=cyclic-import
from redbrick.utils.logging import print_error
from redbrick.common.constants import DEFAULT_URL, MAX_RETRY_ATTEMPTS


class RBClient:
    """Client to communicate with RedBrick AI GraphQL Server."""

    def __init__(self, api_key: str, url: str) -> None:
        """Construct RBClient."""
        self.url = (url or DEFAULT_URL).rstrip("/") + "/graphql/"
        self.session = requests.Session()

        self.api_key = api_key
        assert (
            len(self.api_key) == 43
        ), "Invalid Api Key length, make sure you've copied it correctly"

    def __del__(self) -> None:
        """Garbage collect and close session."""
        self.session.close()

    @property
    def headers(self) -> Dict:
        """Get request headers."""
        return {"RB-SDK-Version": sdk_version, "ApiKey": self.api_key}

    @tenacity.retry(
        reraise=True,
        stop=tenacity.stop_after_attempt(MAX_RETRY_ATTEMPTS),
        wait=tenacity.wait_exponential(multiplier=1, min=1, max=10),
        retry=tenacity.retry_if_not_exception_type(
            (KeyboardInterrupt, PermissionError, ValueError)
        ),
    )
    def execute_query(
        self, query: str, variables: Dict, raise_for_error: bool = True
    ) -> Dict:
        """Execute a graphql query."""
        response = self.session.post(
            self.url,
            headers=self.headers,
            json={"query": query, "variables": variables},
        )
        self._check_status_msg(response.status_code)
        return self._process_json_response(response.json(), raise_for_error)

    @tenacity.retry(
        reraise=True,
        stop=tenacity.stop_after_attempt(MAX_RETRY_ATTEMPTS),
        wait=tenacity.wait_exponential(multiplier=1, min=1, max=10),
        retry=tenacity.retry_if_not_exception_type(
            (KeyboardInterrupt, PermissionError, ValueError)
        ),
    )
    async def execute_query_async(
        self,
        aio_session: aiohttp.ClientSession,
        query: str,
        variables: Dict,
        raise_for_error: bool = True,
    ) -> Dict:
        """Execute a graphql query using asyncio."""
        async with aio_session.post(
            self.url,
            headers=self.headers,
            json={"query": query, "variables": variables},
        ) as response:
            self._check_status_msg(response.status)
            return self._process_json_response(await response.json(), raise_for_error)

    @staticmethod
    def _check_status_msg(response_status: int) -> None:
        if response_status >= 500:
            raise ConnectionError(
                "Internal Server Error: You are probably using an invalid API key"
            )
        if response_status == 403:
            raise PermissionError("Problem authenticating with Api Key")

    @staticmethod
    def _process_json_response(
        response_data: Dict, raise_for_error: bool = True
    ) -> Dict:
        """Process JSON resonse."""
        if "errors" in response_data:
            errors = []
            for error in response_data["errors"]:
                errors.append(error["message"])
                print_error(error["message"])

            if raise_for_error:
                raise ValueError("\n".join(errors))

            del response_data["errors"]

        res = {}
        if "data" in response_data:
            res = response_data["data"]
        else:
            res = response_data
        return res



"""CLI upload command."""
import os
import re
import json
import asyncio
from argparse import ArgumentError, ArgumentParser, Namespace
from typing import List, Dict, Optional

from redbrick.cli.project import CLIProject
from redbrick.cli.cli_base import CLIUploadInterface
from redbrick.common.enums import StorageMethod
from redbrick.utils.logging import print_info, print_warning
from redbrick.utils.files import (
    DICOM_FILE_TYPES,
    IMAGE_FILE_TYPES,
    JSON_FILE_TYPES,
    NIFTI_FILE_TYPES,
    VIDEO_FILE_TYPES,
    find_files_recursive,
)


class CLIUploadController(CLIUploadInterface):
    """CLI upload command controller."""

    def __init__(self, parser: ArgumentParser) -> None:
        """Intialize upload sub commands."""
        parser.add_argument(
            "directory",
            help="The directory containing files to upload to the project",
        )
        parser.add_argument(
            "--as-frames", action="store_true", help="Upload video from image frames"
        )
        parser.add_argument(
            "--as-nifti", action="store_true", help="Upload dicom as nifti files"
        )
        parser.add_argument(
            "--json",
            action="store_true",
            help="Upload json files with list of task objects",
        )
        parser.add_argument(
            "--segment-map", "-m", help="Segmentation mapping file path"
        )
        parser.add_argument(
            "--storage",
            "-s",
            default=self.STORAGE_REDBRICK,
            help="Storage method: "
            + f"({self.STORAGE_REDBRICK} [default], {self.STORAGE_PUBLIC}, <storage id>)",
        )
        parser.add_argument(
            "--ground-truth", action="store_true", help="Upload tasks to ground truth"
        )

    def handler(self, args: Namespace) -> None:
        """Handle upload command."""
        self.args = args
        project = CLIProject.from_path()
        assert project, "Not a valid project"
        self.project = project

        self.handle_upload()

    def handle_upload(self) -> None:  # noqa: ignore=C901
        """Handle empty sub command."""
        # pylint: disable=protected-access, too-many-branches, too-many-locals, too-many-statements
        # pylint: disable=too-many-nested-blocks
        project = self.project.project

        directory = os.path.normpath(self.args.directory)
        if directory.endswith(".json") and os.path.isfile(directory):
            self.args.json = True
        elif not os.path.isdir(directory):
            raise Exception(f"{directory} is not a valid directory")

        data_type = str(project.project_type.value).split("_", 1)[0]

        if self.args.storage == self.STORAGE_REDBRICK:
            storage_id = str(StorageMethod.REDBRICK)
        elif self.args.storage == self.STORAGE_PUBLIC:
            storage_id = str(StorageMethod.PUBLIC)
        elif re.match(
            r"^[\da-f]{8}-([\da-f]{4}-){3}[\da-f]{12}$",
            self.args.storage.strip().lower(),
        ):
            storage_id = str(self.args.storage).strip().lower()
        else:
            raise ArgumentError(None, "")

        file_types = IMAGE_FILE_TYPES
        multiple = False
        if self.args.json:
            file_types = JSON_FILE_TYPES
        elif data_type == "VIDEO":
            if self.args.as_frames:
                multiple = True
            else:
                file_types = VIDEO_FILE_TYPES
        elif data_type == "DICOM":
            if self.args.as_nifti:
                file_types = NIFTI_FILE_TYPES
            else:
                multiple = True
                file_types = DICOM_FILE_TYPES
        elif data_type != "IMAGE":
            raise Exception(
                "Project data type needs to be IMAGE, VIDEO or DICOM to upload files"
            )

        valid_file_types = ", ".join(f".{type}[.gz]" for type in file_types)
        if self.args.json and directory.endswith(".json") and os.path.isfile(directory):
            items_list = [[os.path.abspath(directory)]]
        else:
            print_info(f"Searching for items recursively in {directory}")
            items_list = find_files_recursive(
                directory, set(file_types.keys()), multiple
            )
            if multiple and not items_list:
                print_info(
                    "No items found. Please make sure your directories contain files of the same"
                    + f" type within them, i.e. any one of the following: ({valid_file_types})"
                )
                return

        upload_cache_hash = self.project.conf.get_option("uploads", "cache")
        upload_cache = set(
            self.project.cache.get_data("uploads", upload_cache_hash, True, True) or []
        )

        points: List[Dict] = []
        uploading = set()
        if self.args.json:
            for item_group in items_list:
                with open(item_group[0], "r", encoding="utf-8") as file_:
                    file_data = json.load(file_)
                if not isinstance(file_data, list):
                    print_warning(f"{item_group[0]} is not a list of tasks")
                    continue
                for item in file_data:
                    if (
                        not isinstance(item.get("items"), list)
                        or not item["items"]
                        or not all(isinstance(i, str) for i in item["items"])
                    ):
                        print_warning(f"Invalid {item} in {item_group[0]}")
                        continue

                    if "name" not in item:
                        item["name"] = item["items"][0]
                    if item["name"] in upload_cache or item["name"] in uploading:
                        print_info(f"Skipping duplicate item name: {item['name']}")
                        continue

                    task_dir = os.path.dirname(item_group[0])
                    if "segmentations" in item:
                        if isinstance(item["segmentations"], list):
                            item["segmentations"] = {
                                str(idx): segmentation
                                for idx, segmentation in enumerate(
                                    item["segmentations"]
                                )
                            }
                        if "labelsMap" not in item and isinstance(
                            item["segmentations"], dict
                        ):
                            item["labelsMap"] = [
                                {"labelName": segmentation, "imageIndex": int(idx)}
                                for idx, segmentation in item["segmentations"].items()
                            ]
                        del item["segmentations"]
                    elif "labelsPath" in item:
                        if "labelsMap" not in item:
                            item["labelsMap"] = [
                                {
                                    "labelName": item["labelsPath"],
                                    "imageIndex": 0,
                                }
                            ]
                        del item["labelsPath"]

                    if item.get("labelsMap"):
                        if data_type == "DICOM":
                            for label_map in item["labelsMap"]:
                                if not isinstance(label_map["labelName"], list):
                                    label_map["labelName"] = [label_map["labelName"]]
                                label_map["labelName"] = [
                                    label_name
                                    if os.path.isabs(label_name)
                                    or not os.path.exists(
                                        os.path.join(task_dir, label_name)
                                    )
                                    else os.path.abspath(
                                        os.path.join(task_dir, label_name)
                                    )
                                    for label_name in label_map["labelName"]
                                ]
                                if len(label_map["labelName"]) == 1:
                                    label_map["labelName"] = label_map["labelName"][0]
                        else:
                            del item["labelsMap"]

                    if storage_id != str(StorageMethod.REDBRICK):
                        uploading.add(item["name"])
                        points.append(item)
                        continue

                    for idx, path in enumerate(item["items"]):
                        item_path = (
                            path
                            if os.path.isabs(path)
                            else os.path.join(task_dir, path)
                        )
                        if os.path.isfile(item_path):
                            item["items"][idx] = item_path
                        else:
                            print_warning(f"{path} from {item_group[0]} does not exist")
                            break
                    else:
                        uploading.add(item["name"])
                        points.append(item)
        else:
            for item_group in items_list:
                if multiple:
                    items_dir = os.path.dirname(item_group[0])
                    task_dir = os.path.dirname(items_dir)
                    task_name = os.path.basename(items_dir)
                else:
                    task_dir = os.path.dirname(item_group[0])
                    task_name = os.path.splitext(os.path.basename(item_group[0]))[0]

                label_data: Optional[Dict] = None

                label_file = os.path.join(task_dir, task_name + ".json")
                if os.path.isfile(label_file):
                    with open(label_file, "r", encoding="utf-8") as file_:
                        label_data = json.load(file_)

                item_name = (
                    label_data["name"]
                    if label_data and label_data.get("name")
                    else (
                        re.sub(
                            r"^" + re.escape(directory + os.path.sep) + r"?",
                            "",
                            (
                                os.path.dirname(item_group[0])
                                if multiple
                                else item_group[0]
                            ),
                        )
                        or directory
                    ).replace(os.path.sep, "/")
                )
                if item_name in upload_cache or item_name in uploading:
                    print_info(f"Skipping duplicate item name: {item_name}")
                    continue

                uploading.add(item_name)

                item = {
                    "name": item_name,
                    "items": item_group[:],
                    "labels": label_data.get("labels", []) if label_data else [],
                }
                if label_data and label_data.get("segmentMap"):
                    item["segmentMap"] = label_data["segmentMap"]
                if data_type == "DICOM":
                    if label_data and isinstance(label_data, dict):
                        if (
                            "segmentations" in label_data
                            and "labelsMap" not in label_data
                            and isinstance(label_data["segmentations"], (list, dict))
                        ):
                            if isinstance(label_data["segmentations"], list):
                                label_data["segmentations"] = {
                                    str(idx): segmentation
                                    for idx, segmentation in enumerate(
                                        label_data["segmentations"]
                                    )
                                }
                            if isinstance(label_data["segmentations"], dict):
                                label_data["labelsMap"] = [
                                    {"labelName": segmentation, "imageIndex": int(idx)}
                                    for idx, segmentation in label_data[
                                        "segmentations"
                                    ].items()
                                ]
                        elif (
                            "labelsPath" in label_data and "labelsMap" not in label_data
                        ):
                            label_data["labelsMap"] = [
                                {
                                    "labelName": label_data["labelsPath"],
                                    "imageIndex": 0,
                                }
                            ]
                        if label_data.get("labelsMap"):
                            item["labelsMap"] = []
                            for label_map in label_data["labelsMap"]:
                                if not isinstance(label_map["labelName"], list):
                                    label_map["labelName"] = [label_map["labelName"]]
                                label_map["labelName"] = [
                                    label_name
                                    if os.path.isabs(label_name)
                                    or not os.path.exists(
                                        os.path.join(task_dir, label_name)
                                    )
                                    else os.path.abspath(
                                        os.path.join(task_dir, label_name)
                                    )
                                    for label_name in label_map["labelName"]
                                ]
                                if len(label_map["labelName"]) == 1:
                                    label_map["labelName"] = label_map["labelName"][0]
                                item["labelsMap"].append(
                                    {
                                        "labelName": label_map["labelName"],
                                        "imageIndex": int(label_map["imageIndex"]),
                                        "imageName": label_map.get("imageName"),
                                        "seriesId": label_map.get("seriesId"),
                                    }
                                )
                    elif not multiple and os.path.isdir(
                        os.path.join(task_dir, task_name)
                    ):
                        item["labelsMap"] = [
                            {
                                "labelName": os.path.abspath(
                                    os.path.join(task_dir, task_name)
                                ),
                                "imageIndex": 0,
                            }
                        ]
                    elif multiple and os.path.isfile(
                        os.path.join(task_dir, task_name + ".nii.gz")
                    ):
                        item["labelsMap"] = [
                            {
                                "labelName": os.path.abspath(
                                    os.path.join(task_dir, task_name + ".nii.gz")
                                ),
                                "imageIndex": 0,
                            }
                        ]
                    elif multiple and os.path.isfile(
                        os.path.join(task_dir, task_name + ".nii")
                    ):
                        item["labelsMap"] = [
                            {
                                "labelName": os.path.abspath(
                                    os.path.join(task_dir, task_name + ".nii")
                                ),
                                "imageIndex": 0,
                            }
                        ]

                points.append(item)

        if points:
            print_info(f"Found {len(points)} items")
            loop = asyncio.get_event_loop()
            segmentation_mapping = {}
            if self.args.segment_map:
                with open(self.args.segment_map, "r", encoding="utf-8") as file_:
                    segmentation_mapping = json.load(file_)

            uploads = loop.run_until_complete(
                project.upload._create_tasks(
                    storage_id, points, segmentation_mapping, self.args.ground_truth
                )
            )

            for upload in uploads:
                if upload.get("response"):
                    upload_cache.add(upload["name"])

            self.project.conf.set_option(
                "uploads",
                "cache",
                self.project.cache.set_data("uploads", list(upload_cache), True),
            )
            self.project.conf.save()
        elif not items_list:
            print_info(
                "No items found. Please make sure your directories contain files of valid type,"
                + f" i.e. any one of the following: ({valid_file_types})"
            )



"""Segmentation helper functions."""

from typing import Dict, Any

import numpy as np


def check_mask_map_format(mask: np.ndarray, color_map: Dict) -> Any:
    """
    Validate format of masks.

    Checks the following:
        - If mask dimensions are [m, n, 3]
        - If all the values in mask, are present in color_map
    """
    if len(mask.shape) != 3 or mask.shape[2] != 3:
        raise ValueError(f"Your mask must have shape (m, n, 3), not {mask.shape}")

    if mask.dtype != np.uint8:
        raise ValueError(
            "Your masks must be of type np.uint8, with pixel values between [0, 255]."
        )

    # Creates a map of all the colors
    unique_color_map = np.zeros((len(color_map), 3))
    for i, category in enumerate(color_map):
        unique_color_map[i, :] = color_map[category]

    indexes = np.where(
        (mask[:, :, 0] in unique_color_map[:, 0])
        & (mask[:, :, 1] in unique_color_map[:, 1])
        & (mask[:, :, 2] in unique_color_map[:, 2])
    )
    return indexes



"""Repo for accessing export apis."""
from typing import Optional, List, Dict, Tuple
from datetime import datetime
from dateutil import parser  # type: ignore

import aiohttp

from redbrick.common.export import ExportControllerInterface
from redbrick.common.client import RBClient
from redbrick.repo.shards import TAXONOMY_SHARD, LATEST_TASKDATA_SHARD


class ExportRepo(ExportControllerInterface):
    """Handle API requests to get export data."""

    def __init__(self, client: RBClient) -> None:
        """Construct ExportRepo."""
        self.client = client

    def get_datapoints_input(
        self,
        org_id: str,
        project_id: str,
        first: int = 50,
        presign: bool = False,
        cursor: Optional[str] = None,
    ) -> Tuple[List[Dict], Optional[str]]:
        """Get datapoints that were uploaded to the project."""
        raise NotImplementedError()

    def get_output_info(self, org_id: str, project_id: str) -> Dict:
        """Get info about the output labelset and taxonomy."""
        query_string = (
            """
        query ($orgId: UUID!, $name: String!){
            customGroup(orgId: $orgId, name:$name){
                dataType
                taskType
                datapointCount
                taxonomy {
                    %s
                    colorMap {
                        name
                        color
                        classid
                        trail
                        taskcategory
                    }
                }
            }
        }
        """
            % TAXONOMY_SHARD
        )

        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "name": project_id + "-output",
        }

        result = self.client.execute_query(query_string, query_variables)

        temp: Dict = result["customGroup"]
        return temp

    def get_datapoints_output(
        self,
        org_id: str,
        project_id: str,
        first: int = 50,
        cursor: Optional[str] = None,
    ) -> Tuple[List[Dict], Optional[str]]:
        """Get datapoints that have made it to the output of the project."""
        query_string = """
        query ($orgId: UUID!, $projectId: UUID!, $name: String!,$first: Int, $cursor: String){
        customGroup(orgId: $orgId, name:$name){
            datapointsPaged(first:$first, after:$cursor) {
            entries {
                name
                itemsPresigned:items (presigned:true)
                items(presigned:false)
                task(projectId: $projectId) {
                    taskId
                }
                labelData(customGroupName: $name){
                    createdByEmail
                    labelsData(interpolate: true)
                    labelsMap {
                        labelName
                        imageIndex
                        imageName
                        seriesId
                    }
                }
            }
            cursor
            }
        }
        }
        """
        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "name": project_id + "-output",
            "projectId": project_id,
            "cursor": cursor,
            "first": first,
        }

        result = self.client.execute_query(query_string, query_variables, False)
        custom_group = result.get("customGroup", {}) or {}
        datapoints_paged = custom_group.get("datapointsPaged", {}) or {}
        entries: List[Dict] = datapoints_paged.get("entries", []) or []  # type: ignore
        return entries, datapoints_paged.get("cursor")

    def datapoints_in_project(
        self, org_id: str, project_id: str, stage_name: Optional[str] = None
    ) -> int:
        """Get number of datapoints in project."""
        query_string = """
        query($orgId: UUID!, $projectId: UUID!, $stageName: String, $first: Int) {
            tasksPaged(
                orgId: $orgId
                projectId: $projectId
                stageName: $stageName
                first: $first
            ) {
                count
            }
        }
        """
        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "first": 0,
        }

        result = self.client.execute_query(query_string, query_variables)

        return int(result["tasksPaged"]["count"])

    def get_datapoints_latest(
        self,
        org_id: str,
        project_id: str,
        stage_name: Optional[str] = None,
        cache_time: Optional[datetime] = None,
        first: int = 50,
        cursor: Optional[str] = None,
    ) -> Tuple[List[Dict], Optional[str], Optional[datetime]]:
        """Get the latest datapoints."""
        query_string = (
            """
        query(
            $orgId: UUID!,
            $projectId: UUID!,
            $stageName: String,
            $cacheTime: DateTime,
            $first: Int,
            $cursor: String
        ) {
            tasksPaged(
                orgId: $orgId
                projectId: $projectId
                stageName: $stageName
                cacheTime: $cacheTime
                first: $first
                after: $cursor
            ) {
                entries {
                    taskId
                    dpId
                    currentStageName
                    %s
                }
                cursor
                cacheTime
            }
            }
        """
            % LATEST_TASKDATA_SHARD
        )
        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "cacheTime": None if cache_time is None else cache_time.isoformat(),
            "first": first,
            "cursor": cursor,
        }

        result = self.client.execute_query(query_string, query_variables, False)
        tasks_paged = result.get("tasksPaged", {}) or {}
        entries: List[Dict] = tasks_paged.get("entries", []) or []  # type: ignore
        new_cache_time: Optional[str] = tasks_paged.get("cacheTime")
        return (
            entries,
            tasks_paged.get("cursor"),
            parser.parse(new_cache_time) if new_cache_time else None,
        )

    def get_datapoint_latest(self, org_id: str, project_id: str, task_id: str) -> Dict:
        """Get the latest labels for a single bdatapoint."""
        query_string = (
            """
        query($orgId: UUID!, $projectId: UUID!, $taskId: UUID!) {
            task(
                orgId: $orgId
                projectId: $projectId
                taskId: $taskId
            ) {
                taskId
                currentStageName
                %s
            }
            }
        """
            % LATEST_TASKDATA_SHARD
        )
        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "taskId": task_id,
        }

        result: Dict[str, Dict] = self.client.execute_query(
            query_string, query_variables, False
        )

        return result.get("task", {}) or {}

    async def get_labels(
        self, session: aiohttp.ClientSession, org_id: str, project_id: str, dp_id: str
    ) -> Dict:
        """Get input labels."""
        query = """
        query dataPoint(
            $orgId: UUID!
            $dpId: UUID!
            $name: String!
        ) {
            dataPoint(orgId: $orgId, dpId: $dpId) {
                labelData(customGroupName: $name) {
                    dpId
                    labelsData
                }
            }
        }
        """

        variables = {
            "orgId": org_id,
            "dpId": dp_id,
            "name": f"{project_id}-input",
        }

        response = await self.client.execute_query_async(session, query, variables)
        label_data: Dict = response.get("dataPoint", {}).get("labelData", {})
        return label_data

    def task_search(
        self,
        org_id: str,
        project_id: str,
        stage_name: Optional[str] = None,
        task_search: Optional[str] = None,
        first: int = 50,
        after: Optional[str] = None,
    ) -> Tuple[List[Dict], Optional[str]]:
        """Task search."""
        query_string = """
        query tasksList(
            $orgId: UUID!
            $projectId: UUID!
            $stageName: String
            $taskSearch: String
            $first: Int
            $after: String
        ) {
            genericTasks(
                orgId: $orgId
                projectId: $projectId
                stageName: $stageName
                taskSearch: $taskSearch
                first: $first
                after: $after
            ) {
                entries {
                    taskId
                    datapoint {
                        name
                    }
                    currentStageName
                    createdAt
                }
                cursor
            }
        }
        """

        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "taskSearch": task_search,
            "first": first,
            "after": after,
        }

        result = self.client.execute_query(query_string, query_variables, False)
        generic_tasks = result.get("genericTasks", {}) or {}
        entries: List[Dict] = generic_tasks.get("entries", []) or []  # type: ignore

        return entries, generic_tasks.get("cursor")

    def presign_items(
        self, org_id: str, storage_id: str, items: List[str]
    ) -> List[str]:
        """Presign download items."""
        query = """
        query presignItems(
            $orgId: UUID!
            $storageId: UUID!
            $items: [String!]!
        ) {
            presignItems(orgId: $orgId, storageId: $storageId, items: $items)
        }
        """

        variables = {"orgId": org_id, "storageId": storage_id, "items": items}

        response = self.client.execute_query(query, variables)
        presigned_items: List[str] = response.get("presignItems", [])
        return presigned_items



"""Abstract interface to upload."""

from typing import List, Dict, Optional, Any
from pathlib import Path

import aiohttp

from redbrick.common.client import RBClient
from redbrick.common.upload import UploadControllerInterface


class UploadRepo(UploadControllerInterface):
    """Handle communication with backend relating to uploads."""

    def __init__(self, client: RBClient) -> None:
        """Construct ExportRepo."""
        self.client = client

    def create_datapoint(
        self,
        org_id: str,
        project_id: str,
        storage_id: str,
        name: str,
        items: List[str],
        labels: Optional[List[Dict]],
    ) -> str:
        """
        Create a datapoint and returns its taskId.

        Name must be unique in the project.
        """
        raise NotImplementedError()

    async def create_datapoint_async(
        self,
        aio_client: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        storage_id: str,
        name: str,
        items: List[str],
        labels_data: Optional[str],
        labels_map: Optional[List[Dict]] = None,
        is_ground_truth: bool = False,
    ) -> Dict:
        """
        Create a datapoint and returns its taskId.

        Name must be unique in the project.
        """
        query_string = """
            mutation createDatapointSDK(
                $orgId: UUID!
                $projectId: UUID!
                $items: [String!]!
                $name: String!
                $storageId: UUID!
                $labelsData: String
                $labelsMap: [LabelMapInput!]
                $isGroundTruth: Boolean!
            ) {
                createDatapoint(
                    orgId: $orgId
                    projectId: $projectId
                    items: $items
                    name: $name
                    storageId: $storageId
                    labelsData: $labelsData
                    labelsMap: $labelsMap
                    isGroundTruth: $isGroundTruth
                ) {
                    taskId
                }
            }
        """

        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "items": items,
            "name": name,
            "storageId": storage_id,
            "labelsData": labels_data,
            "labelsMap": labels_map,
            "isGroundTruth": is_ground_truth,
        }
        response = await self.client.execute_query_async(
            aio_client, query_string, query_variables
        )
        return response.get("createDatapoint", {})

    async def item_upload_async(
        self,
        aio_client: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        storage_id: str,
        data_type: str,
        task_type: str,
        file_path: str,
        file_name: str,
        file_type: str,
        is_ground_truth: bool = False,
    ) -> Dict:
        """Upload an item."""
        query_string = """
            mutation itemListUploadSuccessSDK(
                $orgId: UUID!
                $projectId: UUID!
                $payload: ItemsListUploadSuccessInput!
            ) {
                itemListUploadSuccess(
                    orgId: $orgId
                    projectId: $projectId
                    payload: $payload
                ) {
                    ok
                }
            }
        """

        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "payload": {
                "orgId": org_id,
                "projectId": project_id,
                "filePath": file_path,
                "fileName": file_name,
                "fileType": file_type,
                "uploadId": project_id,
                "dataType": data_type,
                "taskType": task_type,
                "storageId": storage_id,
                "isGroundTruth": is_ground_truth,
            },
        }
        response = await self.client.execute_query_async(
            aio_client, query_string, query_variables
        )
        return response.get("itemListUploadSuccess", {})

    def items_upload_presign(
        self, org_id: str, project_id: str, files: List[str], file_type: List[str]
    ) -> List[Dict[Any, Any]]:
        """Return presigned URLs to upload files."""
        query_string = """
            query itemsUploadPresign(
                $orgId:UUID!,
                $projectId: UUID!,
                $files: [String]!,
                $fileType:[String]!
            ){
                itemsUploadPresign(
                    orgId:$orgId,
                    projectId: $projectId,
                    files:$files,
                    fileType:$fileType
                ) {
                    items {
                        presignedUrl,
                        filePath,
                        fileName
                    }
                }
            }
        """

        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "files": files,
            "fileType": file_type,
        }
        result = self.client.execute_query(query_string, query_variables)
        assert isinstance(result["itemsUploadPresign"]["items"], list)
        return result["itemsUploadPresign"]["items"]

    def upload_image(self, org_id: str, project_id: str, file_path: Path) -> str:
        """Upload a local image and add labels."""
        raise NotImplementedError()

    async def upload_image_async(
        self, org_id: str, project_id: str, file_path: Path
    ) -> str:
        """Upload a local image and add labels."""
        raise NotImplementedError()

    def delete_tasks(self, org_id: str, project_id: str, task_ids: List[str]) -> bool:
        """Delete tasks in a project."""
        query_string = """
        mutation($orgId: UUID!, $projectId: UUID!, $taskIds: [UUID!]!) {
            deleteTasks(
                orgId: $orgId
                projectId: $projectId
                taskIds: $taskIds
            ) {
                ok
            }
        }
        """
        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "taskIds": task_ids,
        }

        result: Dict[str, Dict] = self.client.execute_query(
            query_string, query_variables
        )

        return (result.get("deleteTasks", {}) or {}).get("ok", False)



"""Abstract interface to exporting."""

from typing import Optional, List, Dict, Tuple
from abc import ABC, abstractmethod
import aiohttp


class LabelingControllerInterface(ABC):
    """Abstract interface to Labeling APIs."""

    @abstractmethod
    def get_labeling_tasks(
        self, org_id: str, project_id: str, stage_name: str, count: int = 5
    ) -> List[Dict]:
        """Get labeling tasks."""

    @abstractmethod
    async def presign_labels_path(
        self,
        session: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        task_id: str,
        file_type: str,
    ) -> Dict:
        """Presign labels path."""

    @abstractmethod
    async def put_labeling_results(
        self,
        session: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        stage_name: str,
        task_id: str,
        labels_data: str,
        labels_path: Optional[str] = None,
        finished: bool = True,
    ) -> None:
        """Put Labeling results."""

    @abstractmethod
    async def put_review_task_result(
        self,
        session: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        stage_name: str,
        task_id: str,
        review_val: bool,
    ) -> None:
        """Put review result for task."""

    @abstractmethod
    def assign_task(
        self, org_id: str, project_id: str, stage_name: str, task_id: str, email: str
    ) -> None:
        """Assign task to specified email."""

    @abstractmethod
    def get_tasks_queue(
        self,
        org_id: str,
        project_id: str,
        stage_name: str,
        first: int,
        cursor: Optional[str] = None,
    ) -> Tuple[List[Dict], str]:
        """Get task queue."""

    @abstractmethod
    def get_task_queue_count(
        self,
        org_id: str,
        project_id: str,
        stage_name: str,
        for_active_learning: bool = False,
    ) -> int:
        """Get the length of the task queue for showing loading."""

    @abstractmethod
    async def move_task_to_start(
        self,
        session: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        task_id: str,
    ) -> None:
        """Move groundtruth task back to start."""



"""Async utils."""
import asyncio
from typing import Awaitable, Coroutine, List, Tuple, TypeVar, Optional, Iterable
import tqdm.asyncio  # type: ignore

from redbrick.common.constants import MAX_CONCURRENCY

ReturnType = TypeVar("ReturnType")  # pylint: disable=invalid-name


async def gather_with_concurrency(
    max_concurrency: int,
    tasks: Iterable[Awaitable[ReturnType]],
    progress_bar_name: Optional[str] = None,
    keep_progress_bar: bool = True,
) -> List[ReturnType]:
    """Utilizes a Semaphore to limit concurrency to n."""
    if not tasks:
        return []

    max_concurrency = min(max_concurrency, MAX_CONCURRENCY)
    semaphore = asyncio.Semaphore(max_concurrency)

    async def sem_task(task: Awaitable[ReturnType]) -> ReturnType:
        async with semaphore:
            return await task

    coros = [sem_task(task) for task in tasks]
    if not progress_bar_name:
        return await asyncio.gather(*coros)

    async def ordered_coroutine(idx: int, task: Coroutine) -> Tuple[int, ReturnType]:
        """Return the index and result of a task."""
        return idx, await task

    ordered_coros = [ordered_coroutine(idx, task) for idx, task in enumerate(coros)]
    result: List[Tuple[int, ReturnType]] = []

    for coro in tqdm.asyncio.tqdm.as_completed(
        ordered_coros, desc=progress_bar_name, leave=keep_progress_bar
    ):
        temp = await coro
        result.append(temp)

    return [res[1] for res in sorted(result, key=lambda x: x[0])]



"""Dicom/nifti related functions."""
import os
from typing import Dict, List, Optional, Tuple, Union

import nibabel  # type: ignore
import numpy

from redbrick.utils.files import uniquify_path
from redbrick.utils.logging import print_error


def process_nifti_download(
    task: Dict, labels_path: Optional[str]
) -> Optional[Union[str, List[str]]]:
    """Process nifti download file."""
    try:
        if not (
            labels_path
            and os.path.isfile(labels_path)
            and any(
                label.get("dicom", {}).get("groupids")
                for label in task.get("labels", [])
            )
        ):
            return labels_path

        img = nibabel.load(labels_path)

        if not isinstance(img, nibabel.Nifti1Image):
            print_error(f"{labels_path} is not a valid NIfTI1 file.")
            return labels_path

        affine = img.affine
        header = img.header
        data = img.get_fdata()

        dirname = uniquify_path(os.path.splitext(labels_path)[0])
        os.mkdir(dirname)
        files: List[str] = []

        for label in task["labels"]:
            if label.get("dicom"):
                instances: List[int] = [label["dicom"]["instanceid"]] + (
                    label["dicom"].get("groupids", []) or []
                )
                filename = uniquify_path(
                    os.path.join(dirname, f"{label['dicom']['instanceid']}.nii")
                )

                new_img = nibabel.Nifti1Image(
                    numpy.where(
                        numpy.isin(data, instances),  # type: ignore
                        label["dicom"]["instanceid"],
                        0,
                    ),
                    affine,
                    header,
                )
                nibabel.save(new_img, filename)
                files.append(filename)

        return files

    except Exception as error:  # pylint: disable=broad-except
        print_error(f"Failed to process {labels_path}: {error}")
        return labels_path


def process_nifti_upload(
    files: List[str],
) -> Tuple[Optional[str], Dict[int, List[int]]]:
    """Process nifti upload files."""
    # pylint: disable=too-many-locals, too-many-branches
    # pylint: disable=too-many-statements, too-many-return-statements
    if not files or any(
        not isinstance(file_, str) or not os.path.isfile(file_) for file_ in files
    ):
        return None, {}

    if len(files) == 1:
        return files[0], {}

    try:
        instance_map: Dict[int, List[int]] = {}
        reverse_instance_map: Dict[Tuple[int, ...], int] = {}
        base_img = nibabel.load(files[0])

        if not isinstance(base_img, nibabel.Nifti1Image) and not isinstance(
            base_img, nibabel.Nifti2Image
        ):
            return None, {}

        base_data = base_img.get_fdata()

        if base_img.get_data_dtype() != numpy.uint16:
            base_img.set_data_dtype(numpy.uint16)
            base_data = numpy.round(base_data).astype(numpy.uint16)  # type: ignore

        if base_data.ndim != 3:
            return None, {}

        for instance in numpy.unique(base_data):  # type: ignore
            if instance and instance not in instance_map:
                inst = int(instance)
                instance_map[inst] = [inst]
                reverse_instance_map[(inst,)] = inst

        for file_ in files[1:]:
            img = nibabel.load(file_)
            data = img.get_fdata()
            if base_data.shape != data.shape:
                return None, {}

            for instance in numpy.unique(data):  # type: ignore
                if instance and instance not in instance_map:
                    inst = int(instance)
                    instance_map[inst] = [inst]
                    reverse_instance_map[(inst,)] = inst

        group_instances = sorted(
            set(range(1, 65536)) - set(instance_map.keys()), reverse=True
        )

        for file_ in files[1:]:
            img = nibabel.load(file_)
            if not isinstance(img, nibabel.Nifti1Image) and not isinstance(
                img, nibabel.Nifti2Image
            ):
                return None, {}

            data = img.get_fdata()
            if img.get_data_dtype() != numpy.uint16:
                data = numpy.round(data).astype(numpy.uint16)  # type: ignore

            for i, j, k in zip(*numpy.where(numpy.logical_and(base_data, data))):
                sub_instances = tuple(
                    sorted(
                        instance_map[int(base_data[i, j, k])]
                        + instance_map[int(data[i, j, k])]
                    )
                )
                if sub_instances not in reverse_instance_map:
                    if not group_instances:
                        return None, {}
                    reverse_instance_map[sub_instances] = group_instances.pop()
                    instance_map[reverse_instance_map[sub_instances]] = list(
                        sub_instances
                    )
                data[i, j, k] = reverse_instance_map[sub_instances]

            base_data = numpy.where(data, data, base_data)

        base_data = numpy.asarray(base_data, dtype=numpy.uint16)

        if isinstance(base_img, nibabel.Nifti1Image):
            new_img = nibabel.Nifti1Image(base_data, base_img.affine, base_img.header)
        else:
            new_img = nibabel.Nifti2Image(base_data, base_img.affine, base_img.header)

        dirname = os.path.join(os.path.expanduser("~"), ".redbrickai", "temp")
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        filename = uniquify_path(os.path.join(dirname, "label.nii"))
        nibabel.save(new_img, filename)

        group_map: Dict[int, List[int]] = {}
        for sub_ids, group_id in reverse_instance_map.items():
            if len(sub_ids) > 1:
                for sub_id in sub_ids:
                    if sub_id not in group_map:
                        group_map[sub_id] = []
                    group_map[sub_id].append(group_id)

        return (filename, group_map)

    except Exception as error:  # pylint: disable=broad-except
        print_error(error)
        return None, {}



"""Interfaces for RedBrick CLI."""
from abc import ABC, abstractmethod
from typing import Optional
from argparse import Namespace

from redbrick.cli.project import CLIProject
from redbrick.utils.logging import print_warning


class CLIInputParams(ABC):
    """CLI Input params handler."""

    entity: Optional[str]
    error_message: str

    @abstractmethod
    def filtrator(self, entity: str) -> str:
        """Filter input entity."""

    @abstractmethod
    def validator(self, entity: str) -> bool:
        """Validate input entity."""

    @abstractmethod
    def get(self) -> str:
        """Get filtered entity value post validation."""

    def from_args(self) -> Optional[str]:
        """Get value from cli args."""
        if self.entity is None:
            return None

        if self.validator(self.entity):
            return self.filtrator(self.entity)

        print_warning(self.error_message + " : " + self.entity)
        return None


class CLIConfigInterface(ABC):
    """CLI config command interface."""

    args: Namespace
    project: CLIProject

    LIST = "list"
    SET = "set"
    ADD = "add"
    REMOVE = "remove"
    CLEAR = "clear"
    VERIFY = "verify"

    @abstractmethod
    def handler(self, args: Namespace) -> None:
        """Handle config command."""

    @abstractmethod
    def handle_config(self) -> None:
        """Handle empty sub command."""

    @abstractmethod
    def handle_list(self) -> None:
        """Handle list sub command."""

    @abstractmethod
    def handle_set(self) -> None:
        """Handle set sub command."""

    @abstractmethod
    def handle_add(self) -> None:
        """Handle add sub command."""

    @abstractmethod
    def handle_remove(self) -> None:
        """Handle remove sub command."""

    @abstractmethod
    def handle_clear(self) -> None:
        """Handle clear sub command."""

    @abstractmethod
    def handle_verify(self, profile: Optional[str] = None) -> None:
        """Handle verify sub command."""


class CLIInitInterface(ABC):
    """CLI init interface."""

    args: Namespace
    project: CLIProject

    @abstractmethod
    def handler(self, args: Namespace) -> None:
        """Handle init command."""

    @abstractmethod
    def handle_init(self) -> None:
        """Handle empty sub command."""


class CLICloneInterface(ABC):
    """CLI clone interface."""

    args: Namespace
    project: CLIProject

    @abstractmethod
    def handler(self, args: Namespace) -> None:
        """Handle clone command."""

    @abstractmethod
    def handle_clone(self) -> None:
        """Handle empty sub command."""


class CLIInfoInterface(ABC):
    """CLI info interface."""

    args: Namespace
    project: CLIProject

    SETTING_LABELSTORAGE = "labelstorage"

    @abstractmethod
    def handler(self, args: Namespace) -> None:
        """Handle info command."""

    @abstractmethod
    def handle_get(self) -> None:
        """Handle get sub command."""

    @abstractmethod
    def handle_set(self) -> None:
        """Handle set sub command."""

    @abstractmethod
    def handle_info(self) -> None:
        """Handle empty sub command."""


class CLIExportInterface(ABC):
    """CLI export interface."""

    args: Namespace
    project: CLIProject

    TYPE_LATEST = "latest"
    TYPE_GROUNDTRUTH = "groundtruth"

    FORMAT_REDBRICK = "redbrick"
    FORMAT_COCO = "coco"
    FORMAT_MASK = "mask"
    FORMAT_NIFTI = "nifti"

    @abstractmethod
    def handler(self, args: Namespace) -> None:
        """Handle export command."""

    @abstractmethod
    def handle_export(self) -> None:
        """Handle empty sub command."""


class CLIPruneInterface(ABC):
    """CLI prune interface."""

    args: Namespace
    project: CLIProject

    @abstractmethod
    def handler(self, args: Namespace) -> None:
        """Handle prune command."""

    @abstractmethod
    def handle_prune(self) -> None:
        """Handle empty sub command."""


class CLIUploadInterface(ABC):
    """CLI upload interface."""

    args: Namespace
    project: CLIProject

    STORAGE_REDBRICK = "redbrick"
    STORAGE_PUBLIC = "public"

    @abstractmethod
    def handler(self, args: Namespace) -> None:
        """Handle upload command."""

    @abstractmethod
    def handle_upload(self) -> None:
        """Handle empty sub command."""


class CLIInterface(ABC):
    """Main CLI Interface."""

    config: CLIConfigInterface
    init: CLIInitInterface
    clone: CLICloneInterface
    info: CLIInfoInterface
    export: CLIExportInterface
    prune: CLIPruneInterface
    upload: CLIUploadInterface

    CONFIG = "config"
    INIT = "init"
    CLONE = "clone"
    INFO = "info"
    EXPORT = "export"
    PRUNE = "prune"
    UPLOAD = "upload"

    @abstractmethod
    def handle_command(self, args: Namespace) -> None:
        """CLI command main handler."""



"""Enumerations for use across SDK."""

from enum import Enum


class LabelType(Enum):
    """Allowable types for labeling projects.

    - IMAGE_CLASSIFY - Classification of images
    - IMAGE_BBOX - Bounding box detections with images
    - IMAGE_POINT - Keypoint detections with images
    - IMAGE_POLYLINE - Polyline shapes with images
    - IMAGE_POLYGON - Freeform polygon detections with images
    - IMAGE_ELLIPSE - Ellipse shapes with images
    - IMAGE_SEGMENTATION - Pixel level segmentation with images
    - IMAGE_MULTI - All available image labeling types for detection and segmentation

    - VIDEO_BBOX - Bounding box detection and object tracking with video frames
    - VIDEO_POINT - Keypoint detections and object tracking with video frames
    - VIDEO_POLYLINE - Polyline shapes and object tracking with video frames
    - VIDEO_ELLIPSE - Ellipse shapes and object tracking with video frames
    - VIDEO_CLASSIFY - Frame level Classification of video frames.
    - VIDEO_POLYGON - Freeform polygon detections and object tracking with video frames
    - VIDEO_MULTI - All available video labeling types for annotating videos with multiple shapes

    - DOCUMENT_BBOX - Annotate pdf and other documents with work detection and OCR tools

    - DICOM_SEGMENTATION - 3D pixel segmentation of DICOM series data
    """

    IMAGE_ITEMS = "IMAGE_ITEMS"
    IMAGE_CLASSIFY = "IMAGE_CLASSIFY"
    IMAGE_BBOX = "IMAGE_BBOX"
    IMAGE_POINT = "IMAGE_POINT"
    IMAGE_POLYLINE = "IMAGE_POLYLINE"
    IMAGE_POLYGON = "IMAGE_POLYGON"
    IMAGE_ELLIPSE = "IMAGE_ELLIPSE"
    IMAGE_SEGMENTATION = "IMAGE_SEGMENTATION"
    IMAGE_MULTI = "IMAGE_MULTI"

    VIDEO_ITEMS = "VIDEO_ITEMS"
    VIDEO_BBOX = "VIDEO_BBOX"
    VIDEO_POINT = "VIDEO_POINT"
    VIDEO_POLYLINE = "VIDEO_POLYLINE"
    VIDEO_ELLIPSE = "VIDEO_ELLIPSE"
    VIDEO_CLASSIFY = "VIDEO_CLASSIFY"
    VIDEO_POLYGON = "VIDEO_POLYGON"
    VIDEO_MULTI = "VIDEO_MULTI"

    DOCUMENT_BBOX = "DOCUMENT_BBOX"

    DICOM_SEGMENTATION = "DICOM_SEGMENTATION"


class StorageMethod:
    """Special case storage method Ids.

    - PUBLIC - Access files from a public cloud storage service or local storage
    - REDBRICK - Access files from the RedBrickAI servers
    """

    PUBLIC = "11111111-1111-1111-1111-111111111111"
    REDBRICK = "22222222-2222-2222-2222-222222222222"


class TaskStates(Enum):
    """Potential states of task status."""

    UNASSIGNED = "UNASSIGNED"
    ASSIGNED = "ASSIGNED"
    IN_PROGRESS = "IN_PROGRESS"
    COMPLETED = "COMPLETED"
    PROBLEM = "PROBLEM"



"""Main file for converting RedBrick format to coco format."""
import asyncio
from typing import Dict, List, Optional, Tuple

import aiohttp
from yarl import URL
import tenacity

from redbrick.utils.async_utils import gather_with_concurrency
from redbrick.utils import aioimgspy
from redbrick.utils.logging import print_warning
from redbrick.common.constants import MAX_CONCURRENCY, MAX_RETRY_ATTEMPTS
from .polygon import rb2coco_polygon
from .bbox import rb2coco_bbox
from .categories import rb_get_class_id, rb2coco_categories_format


async def _get_image_dimension_map(
    datapoints: List[Dict],
) -> Dict[str, Tuple[int, int]]:
    """Get a map from taskId to (width, height) of the images."""

    @tenacity.retry(
        reraise=True,
        stop=tenacity.stop_after_attempt(MAX_RETRY_ATTEMPTS),
        wait=tenacity.wait_exponential(multiplier=1, min=1, max=10),
        retry=tenacity.retry_if_not_exception_type(
            (KeyboardInterrupt, PermissionError, ValueError)
        ),
    )
    async def _get_size(
        session: aiohttp.ClientSession, datapoint: Dict
    ) -> Tuple[str, Tuple[int, int]]:
        if not datapoint["itemsPresigned"] or not datapoint["itemsPresigned"][0]:
            return datapoint["taskId"], (0, 0)
        async with session.get(
            # encode with yarl so that aiohttp doesn't encode again.
            URL(datapoint["itemsPresigned"][0], encoded=True)
        ) as response:
            temp = await aioimgspy.probe(response.content)  # type: ignore
        return datapoint["taskId"], (temp["width"], temp["height"])

    # limit to 30, default is 100, cleanup is done by session
    conn = aiohttp.TCPConnector(limit=MAX_CONCURRENCY)
    async with aiohttp.ClientSession(connector=conn) as session:
        coros = [_get_size(session, dpoint) for dpoint in datapoints]
        all_sizes = await gather_with_concurrency(10, coros, "Getting image dimensions")

    await asyncio.sleep(0.250)  # give time to close ssl connections
    return {temp[0]: temp[1] for temp in all_sizes}


# pylint: disable=too-many-locals
def coco_converter(
    datapoints: List[Dict],
    taxonomy: Dict,
    image_dims_map: Optional[Dict[str, Tuple[int, int]]] = None,
) -> Dict:
    """Convert redbrick labels to standard coco format."""
    coco_categories = rb2coco_categories_format(taxonomy)

    if image_dims_map is None:
        loop = asyncio.get_event_loop()
        image_dims_map = loop.run_until_complete(_get_image_dimension_map(datapoints))

    images: List[Dict] = []
    annotations: List[Dict] = []
    for data in datapoints:
        file_name = data["name"]
        task_id = data["taskId"]
        labels = data["labels"]

        width, height = image_dims_map[task_id]

        current_image_id = len(images)
        image_entry = {
            "id": current_image_id,
            "task_id": task_id,
            "file_name": file_name,
            "raw_url": data["items"][0],
            "height": height,
            "width": width,
        }
        if "itemsPresigned" in data:
            image_entry["signed_url"] = data["itemsPresigned"][0]

        images.append(image_entry)

        skipped_labels = 0
        for label in labels:
            annotation_index = len(annotations)
            try:
                if label.get("bbox2d"):
                    class_id = rb_get_class_id(label["category"][0], taxonomy)
                    coco_label = rb2coco_bbox(
                        label,
                        annotation_index,
                        current_image_id,
                        class_id,
                        width,
                        height,
                    )
                    annotations.append(coco_label)
                elif label.get("polygon"):
                    class_id = rb_get_class_id(label["category"][0], taxonomy)
                    coco_label = rb2coco_polygon(
                        label,
                        annotation_index,
                        current_image_id,
                        class_id,
                        width,
                        height,
                    )
                    annotations.append(coco_label)
                else:
                    skipped_labels += 1
            except Exception:  # pylint: disable=broad-except
                skipped_labels += 1

        if skipped_labels:
            if labels == skipped_labels:
                print_warning(
                    f"No bbox/polygon labels found for task {data['taskId']}, skipping"
                )
            else:
                print_warning(
                    f"Skipped {skipped_labels} non bbox/polygon labels for {data['taskId']}"
                )

    return {
        "images": images,
        "annotations": annotations,
        "categories": coco_categories,
        "info": {},
        "licenses": [],
    }



"""CLI configuration handler."""
import os
from configparser import ConfigParser
from typing import Dict, Optional


class CLIConfiguration:
    """CLIConfiguration entity."""

    _conf_file: str
    _conf: ConfigParser

    def __init__(self, conf_file: str) -> None:
        """Initialize CLIConfiguration."""
        self._conf_file = conf_file
        self._conf = ConfigParser()

        if self.exists:
            self._conf.read(self._conf_file)

    @property
    def exists(self) -> bool:
        """Boolean flag to indicate if configuration file exists."""
        if os.path.exists(self._conf_file):
            if os.path.isfile(self._conf_file):
                return True
            raise Exception(f"Not a file {self._conf_file}")
        return False

    def get_section(
        self, section: str, default: Optional[Dict[str, str]] = None
    ) -> Optional[Dict[str, str]]:
        """Get section from configuration."""
        if self.exists and section in self._conf:
            return dict(self._conf[section].items())
        return default

    def set_section(self, section: str, value: Dict[str, str]) -> None:
        """Set section into configuration."""
        self._conf[section] = value

    def get_option(
        self, section: str, option: str, default: Optional[str] = None
    ) -> Optional[str]:
        """Get option from configuration."""
        if self.exists and section in self._conf and option in self._conf[section]:
            return self._conf[section][option]
        return default

    def set_option(self, section: str, option: str, value: str) -> None:
        """Set option into configuration."""
        if section not in self._conf:
            self._conf[section] = {}
        self._conf[section][option] = value

    def save(self) -> None:
        """Save configuration into file."""
        assert os.path.isdir(os.path.dirname(self._conf_file)), "Not a valid project"
        with open(self._conf_file, "w", encoding="utf-8") as conf:
            self._conf.write(conf)



"""Tests for upload module."""
import json
import sys
from typing import List, Dict

from uuid import uuid4
import matplotlib.pyplot as plt  # type: ignore
import pytest

from redbrick.common.enums import StorageMethod
from redbrick.utils.files import get_file_type
from redbrick.upload.public import Upload
from redbrick.tests.conftest import *  # pylint: disable=wildcard-import, unused-wildcard-import


def test_mask_rbai() -> None:
    """Test mask to polygon conversion."""
    if sys.platform not in ("linux", "darwin"):
        # Don't run this test for windows, w/o rasterio will fail.
        assert True
        return
    # read mask that was generated by rbai export
    # i.e. rbai polygon -> png mask format
    mask = plt.imread(
        "redbrick/upload/mask_test/88664c8f-c6f8-4d5a-918e-41d8441a4509.png"
    )
    with open(
        "redbrick/upload/mask_test/class_map.json", "r", encoding="utf-8"
    ) as file:
        class_map = json.load(file)

    # convert the mask to rbai polygon
    dp_entry = Upload.mask_to_rbai(mask, class_map, "items123", "name123")

    # read original rbai polygon format
    with open(
        "redbrick/upload/mask_test/redbrick_export_segment.json", "r", encoding="utf-8"
    ) as file:
        datapoints = json.load(file)
    dp_entry_real = datapoints[0]

    # compare the two polygon formats to confirm mask -> polygon is correct
    for label_real in dp_entry_real["labels"]:
        for label in dp_entry["labels"]:
            if label_real["category"] == label["category"]:
                assert (
                    label_real["pixel"]["regions"].sort()
                    == label["pixel"]["regions"].sort()
                )

                if label_real["pixel"]["holes"]:
                    assert (
                        label_real["pixel"]["holes"].sort()
                        == label["pixel"]["holes"].sort()
                    )


def test_file_type_extraction() -> None:
    """Tests the extraction of MIME file type."""
    filepath = "folder/subfolder/image.png"
    filetype = get_file_type(filepath)
    assert filetype[0] == "png" and filetype[1] == "image/png"


def test_file_type_extraction_invalid() -> None:
    """Check invalid file extraction."""
    filepath = "folder/subfolder/image.csv"
    try:
        get_file_type(filepath)
        assert False
    except ValueError as error:
        assert type(error).__name__ == "ValueError"


def test_items_validity() -> None:
    """Check invalid items."""
    # pylint: disable=protected-access
    file_name = f"{uuid4()}.png"
    with open(file_name, "w", encoding="utf-8") as file_:
        file_.write("test")

    items = [2, "", "missing.txt", file_name]
    invalid = Upload._check_validity_of_items(items)  # type: ignore
    os.remove(file_name)
    assert len(invalid) == len(items) - 1


@pytest.mark.slow
def test_invalid_upload_object(project: RBProject) -> None:
    """Check invalid item list upload."""
    tasks: List[Dict] = [
        {},
        {"name": None},
        {"items": []},
        {"name": "test"},
        {"name": "test", "items": ""},
        {"name": "test", "items": []},
        {"name": "test", "items": [0]},
        {
            "name": "test",
            "items": [
                "https://datasets.redbrickai.com/bccd/BloodImage_00000.jpg",
                "https://datasets.redbrickai.com/bccd/BloodImage_00000.jpg",
            ],
        },
        {
            "name": "test",
            "items": ["https://datasets.redbrickai.com/bccd/BloodImage_00000.jpg"],
            "labels": {},
        },
        {
            "name": str(uuid4()),
            "items": ["https://datasets.redbrickai.com/bccd/BloodImage_00000.jpg"],
        },
    ]
    task_objects = project.upload.create_datapoints(StorageMethod.PUBLIC, tasks)
    assert (
        len(task_objects) == len(tasks)
        and len(list(filter(lambda task: "response" in task, task_objects))) == 1
    )



"""Input text handler."""
from typing import Optional

from InquirerPy import inquirer  # type: ignore

from redbrick.cli.cli_base import CLIInputParams


class CLIInputText(CLIInputParams):
    """Input text handler."""

    def __init__(
        self,
        entity: Optional[str],
        name: str,
        default: str = "",
        allow_empty: bool = False,
    ) -> None:
        """Init handlers."""
        self.entity = entity
        self.error_message = "Empty value"
        self.name = name
        self.default = default
        self.allow_empty = allow_empty

    def filtrator(self, entity: str) -> str:
        """Filter input entity."""
        return entity.strip()

    def validator(self, entity: str) -> bool:
        """Validate input entity."""
        text = self.filtrator(entity)
        return True if self.allow_empty else bool(text)

    def get(self) -> str:
        """Get filtered text value post validation."""
        self.entity = self.from_args()
        if self.entity is None:
            self.entity = inquirer.text(
                qmark=">",
                amark=">",
                message=self.name + ":",
                default=self.default,
                transformer=self.filtrator,
                filter=self.filtrator,
                validate=self.validator,
                invalid_message=self.error_message,
            ).execute()
        return self.entity



"""CLI config command."""
import os
from argparse import ArgumentError, ArgumentParser, Namespace
from typing import Optional
import functools

from InquirerPy import inquirer  # type: ignore
from InquirerPy.utils import color_print  # type: ignore
from InquirerPy.separator import Separator  # type: ignore
from halo import Halo  # type: ignore

from redbrick import _populate_context
from redbrick.cli.input import (
    CLIInputAPIKey,
    CLIInputUUID,
    CLIInputURL,
    CLIInputProfile,
)
from redbrick.cli.project import CLIProject
from redbrick.cli.cli_base import CLIConfigInterface
from redbrick.common.context import RBContext
from redbrick.organization import RBOrganization


class CLIConfigController(CLIConfigInterface):
    """CLI config command controller."""

    def __init__(self, parser: ArgumentParser) -> None:
        """Intialize config sub commands."""
        sub_command = parser.add_subparsers(
            title="Config sub-commands", dest="sub_command"
        )

        parser.add_argument("--org", "-o", help="Org ID")
        parser.add_argument("--key", "-k", help="API Key")
        parser.add_argument("--url", "-u", help="Endpoint URL")
        parser.add_argument("--profile", "-p", help="Profile name")
        parser.add_argument(
            "--force", "-f", action="store_true", help="Force create new credentials"
        )

        # list_sub_command
        _ = sub_command.add_parser(self.LIST, help="List all profiles")

        set_sub_command = sub_command.add_parser(self.SET, help="Set default profile")
        set_sub_command.add_argument("profile", nargs="?", help="Profile name")

        add_sub_command = sub_command.add_parser(self.ADD, help="Add a new profile")
        add_sub_command.add_argument("--org", "-o", help="Org ID")
        add_sub_command.add_argument("--key", "-k", help="API Key")
        add_sub_command.add_argument("--url", "-u", help="Endpoint URL")
        add_sub_command.add_argument("--profile", "-p", help="Profile name")

        unset_sub_command = sub_command.add_parser(self.REMOVE, help="Remove a profile")
        unset_sub_command.add_argument("profile", nargs="?", help="Profile name")

        # clear_sub_command
        _ = sub_command.add_parser(self.CLEAR, help="Clear all credentials")

        verify_sub_command = sub_command.add_parser(
            self.VERIFY, help="Verify a profile"
        )
        verify_sub_command.add_argument("profile", nargs="?", help="Profile name")

    def handler(self, args: Namespace) -> None:
        """Handle config command."""
        self.args = args
        self.project = CLIProject(required=False)

        if not args.sub_command:
            self.handle_config()
        elif args.sub_command == self.LIST:
            self.handle_list()
        elif args.sub_command == self.SET:
            self.handle_set()
        elif args.sub_command == self.ADD:
            self.handle_add()
        elif args.sub_command == self.REMOVE:
            self.handle_remove()
        elif args.sub_command == self.CLEAR:
            self.handle_clear()
        elif args.sub_command == self.VERIFY:
            self.handle_verify()
        else:
            raise ArgumentError(None, "")

    def handle_config(self) -> None:
        """Handle empty sub command."""
        if self.project.creds.exists:
            if not self.args.force:
                confirmation = inquirer.confirm(
                    message="Credentials file already exists. Overwrite?", default=False
                ).execute()
                if not confirmation:
                    return
            self.project.creds.remove()

        self.handle_add()

    def handle_list(self) -> None:
        """Handle list sub command."""
        assert self.project.creds.exists, "Credentials file does not exist"
        default_profile = self.project.creds.selected_profile
        profiles = self.project.creds.profile_names
        data = []
        max_length = 0
        for profile in profiles:
            color = "green" if profile == default_profile else ""
            cur_data = [
                (color, f"{'*' if profile == default_profile else ''}{profile}\n"),
                (color, f"{Separator('-'*len(profile)*2)}\n"),
            ]
            for key, value in self.project.creds.get_profile(profile).items():
                value = (
                    ("***" + value[-3:])
                    if key == "key" and not os.environ.get("REDBRICK_SDK_DEBUG")
                    else value
                )
                info = f"{key}={value}"
                max_length = max(max_length, len(info))
                cur_data.append((color, f"{info}\n"))
            data.append(cur_data)
        if data:
            color_print(
                functools.reduce(
                    lambda x, y: x + [("", f"{Separator('*'*max_length)}\n")] + y, data
                )
            )

    def handle_set(self) -> None:
        """Handle set sub command."""
        assert self.project.creds.exists, "Credentials file does not exist"
        profile = CLIInputProfile(
            self.args.profile, self.project.creds.profile_names
        ).get()
        self.project.creds.set_default(profile)
        self.project.creds.save()

    def handle_add(self) -> None:
        """Handle add sub command."""
        org_id = CLIInputUUID(self.args.org, "Org ID").get()
        api_key = CLIInputAPIKey(self.args.key).get()
        url = CLIInputURL(self.args.url).get()
        profile = CLIInputProfile(
            self.args.profile, self.project.creds.profile_names, True
        ).get()

        self.project.creds.add_profile(profile, api_key, org_id, url)
        self.handle_verify(profile)

        self.project.creds.set_default(profile)
        self.project.creds.save()

    def handle_remove(self) -> None:
        """Handle remove sub command."""
        assert self.project.creds.exists, "Credentials file does not exist"
        profile = CLIInputProfile(
            self.args.profile, self.project.creds.profile_names
        ).get()

        self.project.creds.remove_profile(profile)
        self.project.creds.save()

    def handle_clear(self) -> None:
        """Handle clear sub command."""
        self.project.creds.remove()

    def handle_verify(self, profile: Optional[str] = None) -> None:
        """Handle verify sub command."""
        if profile is None:
            selected_profile = None
            try:
                selected_profile = self.project.creds.selected_profile
            except AssertionError:
                pass
            profile = CLIInputProfile(
                self.args.profile,
                self.project.creds.profile_names,
                default=selected_profile,
            ).get()

        profile_details = self.project.creds.get_profile(profile)
        context = _populate_context(
            RBContext(
                api_key=profile_details["key"].strip(),
                url=profile_details["url"].strip().rstrip("/"),
            )
        )

        with Halo(text="Fetching organization", spinner="dots") as spinner:
            try:
                org = RBOrganization(context, profile_details["org"])
                spinner.succeed(str(org))
            except Exception as error:
                spinner.fail()
                raise error



"""Organization class."""
from datetime import datetime
from functools import partial
from typing import List, Optional, Dict, Union
from tqdm import tqdm  # type: ignore

from redbrick.common.enums import LabelType
from redbrick.common.context import RBContext
from redbrick.project import RBProject
from redbrick.utils.logging import print_info, print_warning, handle_exception
from redbrick.utils.pagination import PaginationIterator
from .basic_project import get_active_learning_project, get_basic_project


class RBOrganization:
    """
    Representation of RedBrick organization.

    The RBOrganization object allows you to programmatically interact with
    your RedBrick organization. This class provides methods for querying your
    organization and doing other high level actions.
    """

    def __init__(self, context: RBContext, org_id: str) -> None:
        """Construct RBOrganization."""
        self.context = context

        self._org_id = org_id
        self._name: str

        self._get_org()

    def _get_org(self) -> None:
        org = self.context.project.get_org(self._org_id)
        self._name = org["name"]

    @handle_exception
    def taxonomies(self, only_name: bool = True) -> Union[List[str], List[Dict]]:
        """Get a list of taxonomy names/objects in the organization."""
        taxonomies = self.context.project.get_taxonomies(self._org_id)
        if only_name:
            return [tax["name"] for tax in taxonomies]
        return taxonomies

    def _all_projects_raw(self) -> List[Dict]:
        """Get and filter entries from server."""
        projects = self.context.project.get_projects(self._org_id)
        projects = list(filter(lambda x: x["status"] == "CREATION_SUCCESS", projects))

        return projects

    @handle_exception
    def projects(self) -> List[RBProject]:
        """Get a list of active projects in the organization."""
        projects = self._all_projects_raw()
        return [
            RBProject(self.context, self._org_id, proj["projectId"])
            for proj in tqdm(projects)
        ]

    @property
    def org_id(self) -> str:
        """Retrieve the unique org_id of this organization."""
        return self._org_id

    @property
    def name(self) -> str:
        """Retrieve unique name of this organization."""
        return self._name

    def __str__(self) -> str:
        """Get string representation of RBOrganization object."""
        return f"RedBrick AI Organization - {self._name} - ( {self._org_id} )"

    def __repr__(self) -> str:
        """Representation of object."""
        return str(self)

    @handle_exception
    def create_project(
        self,
        name: str,
        label_type: LabelType,
        taxonomy_name: str,
        reviews: int = 0,
        exists_okay: bool = False,
        active_learning: Optional[Dict] = None,
    ) -> RBProject:
        """
        Create a project within the organization.

        This method creates an organization in a similar fashion to the
        quickstart on the RedBrick Ai create project page.

        Parameters
        --------------
        name: str
            A unique name for your project

        label_type: redbrick.LabelType
            Configures the label and data type of your project.

        taxonomy_name: str
            The name of the taxonomy you want to use for this project.
            Taxonomies can be found on the left side bar of the platform.

        reviews: int = 0
            The number of review stages that you want to add after the label
            stage.

        exists_okay: bool = False
            Allow projects with the same name to be returned instead of trying to create
            a new project. Useful for when running the same script repeatedly when you
            do not want to keep creating new projects.

        Returns
        --------------
        redbrick.project.RBProject
            A RedBrick Project object.
        """
        if active_learning is not None:
            print_warning("active_learning arg is deprecated and will be ignored")
        stages = get_basic_project(reviews)

        if exists_okay:
            print_info("exists_okay=True... checking for project with same name")
            all_projects = self._all_projects_raw()
            same_name = list(filter(lambda x: x["name"] == name, all_projects))
            if same_name:
                temp = RBProject(self.context, self.org_id, same_name[0]["projectId"])
                if temp.project_type != label_type:
                    raise ValueError(
                        "Project with matching name exists, but it has a different type"
                    )
                if temp.taxonomy_name != taxonomy_name:
                    raise ValueError(
                        "Project with matching name exists, but it has a different taxonomy"
                    )

                print_warning(
                    "exists_okay=True... returning project that already existed"
                )
                return temp

        try:
            project_data = self.context.project.create_project(
                self.org_id, name, stages, label_type.value, taxonomy_name
            )
        except ValueError as error:
            raise Exception(
                "Project with same name exists, try setting exists_okay=True to"
                + " return this project instead of creating a new one"
            ) from error

        return RBProject(self.context, self.org_id, project_data["projectId"])

    @handle_exception
    def labeling_time(
        self, start_date: datetime, end_date: datetime, concurrency: int = 50
    ) -> List[Dict]:
        """Get information of tasks labeled between two dates (both inclusive)."""
        tasks: List[Dict]
        my_iter = PaginationIterator(
            partial(
                self.context.project.get_labeling_information,
                self._org_id,
                start_date,
                end_date,
                concurrency,
            )
        )
        with tqdm(my_iter, unit=" tasks") as progress:
            tasks = [
                {
                    "projectId": task["project"]["projectId"],
                    "taskId": task["taskId"],
                    "completedBy": task["user"]["email"],
                    "timeSpent": task["timeSpent"],
                    "completedAt": task["date"],
                }
                for task in progress
            ]

        return tasks

    @handle_exception
    def create_taxonomy(
        self,
        name: str,
        categories: List[Dict],
        attributes: Optional[List[Dict]] = None,
        task_categories: Optional[List[Dict]] = None,
        task_attributes: Optional[List[Dict]] = None,
    ) -> None:
        """Create Taxonomy.

        Format reference for categories and attributes objects:
        https://docs.redbrickai.com/python-sdk/sdk-overview/reference#taxonomy-objects
        """
        if self.context.project.create_taxonomy(
            self.org_id,
            name,
            [{"name": "object", "children": categories}],
            attributes,
            [{"name": "object", "children": task_categories}]
            if task_categories
            else task_categories,
            task_attributes,
        ):
            print_info(f"Successfully created taxonomy: {name}")



"""Handler for file upload/download."""
import os
import gzip
from typing import List, Optional, Tuple, Set

import asyncio
import aiohttp
import aiofiles  # type: ignore
from yarl import URL
import tenacity
from natsort import natsorted, ns

from redbrick.common.constants import MAX_CONCURRENCY, MAX_RETRY_ATTEMPTS
from redbrick.utils.async_utils import gather_with_concurrency

IMAGE_FILE_TYPES = {
    "png": "image/png",
    "jpg": "image/jpeg",
    "jpeg": "image/jpeg",
    "bmp": "image/bmp",
    "dcm": "application/dicom",
}

VIDEO_FILE_TYPES = {
    "mp4": "video/mp4",
    "avi": "video/x-msvideo",
    "dcm": "application/dicom",
}

DICOM_FILE_TYPES = {"dcm": "application/dicom", "ima": "application/dicom"}

JSON_FILE_TYPES = {"json": "application/json"}

NIFTI_FILE_TYPES = {"nii": "application/octet-stream"}

FILE_TYPES = {
    **IMAGE_FILE_TYPES,
    **VIDEO_FILE_TYPES,
    **DICOM_FILE_TYPES,
    **JSON_FILE_TYPES,
    **NIFTI_FILE_TYPES,
}


def get_file_type(file_path: str) -> Tuple[str, str]:
    """
    Return file type.

    Return
    ------------
    [file_ext, file_type]
        file_ext: png, jpeg, jpg etc.
        file_type: this is the MIME file type e.g. image/png
    """
    file_ext = file_path.split("?", 1)[0].rstrip("/").lower()
    if file_ext.endswith(".gz"):
        file_ext = file_ext[:-3]
    file_ext = file_ext.rsplit(".", 1)[-1].lower()

    if file_ext not in FILE_TYPES:
        raise ValueError(
            f"Unsupported file type {file_ext}! Only {','.join(FILE_TYPES.keys())} are supported"
        )
    return file_ext, FILE_TYPES[file_ext]


def find_files_recursive(
    root: str, file_types: Set[str], multiple: bool = False
) -> List[List[str]]:
    """Find files recursively in a directory, that belong to a list of allowed file types."""
    if not os.path.isdir(root):
        return []

    items = []
    list_items = []
    discard_list_items = False

    for item in os.listdir(root):
        if item.startswith("."):
            continue
        path = os.path.join(root, item)
        if os.path.isdir(path):
            items.extend(find_files_recursive(path, file_types, multiple))
            discard_list_items = True
        elif os.path.isfile(path) and (
            item.rsplit(".", 1)[-1].lower() in file_types
            or (
                "." in item
                and item.rsplit(".", 1)[-1].lower() == "gz"
                and item.rsplit(".", 2)[-2].lower() in file_types
            )
        ):
            if multiple:
                if not discard_list_items:
                    list_items.append(path)
            else:
                items.append([path])
        else:
            discard_list_items = True

    if (
        not discard_list_items
        and len({item.rsplit(".", 1)[-1].lower() for item in list_items}) == 1
    ):
        items.append(list(natsorted(list_items, alg=ns.IGNORECASE)))  # type: ignore

    return items


def uniquify_path(path: str) -> str:
    """Provide unique path with number index."""
    filename, extension = os.path.splitext(path)
    counter = 1

    while os.path.exists(path):
        path = filename + " (" + str(counter) + ")" + extension
        counter += 1

    return path


async def upload_files(
    files: List[Tuple[str, str, str]],
    progress_bar_name: Optional[str] = "Uploading files",
    keep_progress_bar: bool = True,
) -> List[bool]:
    """Upload files from local path to url (file path, presigned url, file type)."""

    @tenacity.retry(
        reraise=True,
        stop=tenacity.stop_after_attempt(MAX_RETRY_ATTEMPTS),
        wait=tenacity.wait_exponential(multiplier=1, min=1, max=10),
        retry=tenacity.retry_if_not_exception_type((KeyboardInterrupt,)),
        retry_error_callback=lambda _: False,
    )
    async def _upload_file(
        session: aiohttp.ClientSession, path: str, url: str, file_type: str
    ) -> bool:
        if not path or not url or not file_type:
            return False
        async with aiofiles.open(path, mode="rb") as file_:  # type: ignore
            data = await file_.read()

        async with session.put(
            url,
            headers={"Content-Type": file_type, "Content-Encoding": "gzip"},
            data=data if path.endswith(".gz") else gzip.compress(data),
        ) as response:
            if response.status == 200:
                return True
            raise ConnectionError(f"Error in uploading {path} to RedBrick")

    # limit to 30, default is 100, cleanup is done by session
    conn = aiohttp.TCPConnector(limit=MAX_CONCURRENCY)
    async with aiohttp.ClientSession(connector=conn) as session:
        coros = [
            _upload_file(session, path, url, file_type)
            for path, url, file_type in files
        ]
        uploaded = await gather_with_concurrency(
            10, coros, progress_bar_name, keep_progress_bar
        )

    await asyncio.sleep(0.250)  # give time to close ssl connections
    return uploaded


async def download_files(
    files: List[Tuple[str, str]],
    progress_bar_name: Optional[str] = "Downloading files",
    keep_progress_bar: bool = True,
) -> List[Optional[str]]:
    """Download files from url to local path (presigned url, file path)."""

    @tenacity.retry(
        reraise=True,
        stop=tenacity.stop_after_attempt(MAX_RETRY_ATTEMPTS),
        wait=tenacity.wait_exponential(multiplier=1, min=1, max=10),
        retry=tenacity.retry_if_not_exception_type(
            (KeyboardInterrupt, PermissionError, ValueError)
        ),
    )
    async def _download_file(
        session: aiohttp.ClientSession, url: str, path: str
    ) -> Optional[str]:
        # pylint: disable=no-member
        if not url or not path:
            return None
        async with session.get(URL(url, encoded=True)) as response:
            if response.status == 200:
                data = await response.read()
                if response.headers.get("Content-Encoding") == "gzip":
                    try:
                        data = gzip.decompress(data)
                    except Exception:  # pylint: disable=broad-except
                        pass
                path = uniquify_path(path)
                async with aiofiles.open(path, "wb") as file_:  # type: ignore
                    await file_.write(data)
                return path
            return None

    # limit to 30, default is 100, cleanup is done by session
    conn = aiohttp.TCPConnector(limit=MAX_CONCURRENCY)
    async with aiohttp.ClientSession(connector=conn) as session:
        coros = [_download_file(session, url, path) for url, path in files]
        paths = await gather_with_concurrency(
            10, coros, progress_bar_name, keep_progress_bar
        )

    await asyncio.sleep(0.250)  # give time to close ssl connections
    return paths



"""Handlers to access APIs for getting projects."""
import json
from typing import List, Dict, Tuple, Optional
from datetime import datetime

from redbrick.common.client import RBClient
from redbrick.common.project import ProjectRepoInterface
from redbrick.repo.shards import TAXONOMY_SHARD


class ProjectRepo(ProjectRepoInterface):
    """Class to manage interaction with project APIs."""

    def __init__(self, client: RBClient) -> None:
        """Construct ProjectRepo."""
        self.client = client

    def get_project(self, org_id: str, project_id: str) -> Dict:
        """
        Get project name and status.

        Raise an exception if project does not exist.
        """
        query = """
            query sdkGetProjectName($orgId: UUID!, $projectId: UUID!){
                project(orgId: $orgId, projectId: $projectId){
                    orgId
                    projectId
                    name
                    status
                    tdType
                    taxonomy {
                        name
                    }
                    projectUrl
                }
            }
        """
        variables = {"orgId": org_id, "projectId": project_id}
        response: Dict[str, Dict] = self.client.execute_query(query, variables)
        if response.get("project"):
            return response["project"]

        raise Exception("Project does not exist")

    def get_stages(self, org_id: str, project_id: str) -> List[Dict]:
        """Get stages."""
        query = """
            query sdkGetStages($orgId: UUID!, $projectId: UUID!){
                stages(orgId: $orgId, projectId: $projectId){
                    stageName
                    brickName
                    stageConfig
                }
            }
        """
        variables = {"orgId": org_id, "projectId": project_id}
        response: Dict[str, List[Dict]] = self.client.execute_query(query, variables)
        return response["stages"]

    def create_project(
        self, org_id: str, name: str, stages: List[dict], td_type: str, tax_name: str
    ) -> Dict:
        """Create a project and return project_id."""
        query = """
            mutation createProjectSimple(
                $orgId: UUID!
                $name: String!
                $stages: [StageInputSimple!]!
                $tdType: TaskDataType!
                $taxonomyName: String!
                $taxonomyVersion: Int!
            ) {
                createProjectSimple(
                orgId: $orgId
                name: $name
                stages: $stages
                tdType: $tdType
                taxonomyName: $taxonomyName
                taxonomyVersion: $taxonomyVersion
                ) {
                ok
                errors
                project {
                    projectId
                    name
                    desc
                    projectUrl
                }
                stages {
                    stageName
                    brickName
                }
                }
            }
        """
        for stage in stages:
            stage["stageConfig"] = json.dumps(stage["stageConfig"])
        variables = {
            "orgId": org_id,
            "name": name,
            "stages": stages,
            "tdType": td_type,
            "taxonomyName": tax_name,
            "taxonomyVersion": 1,
        }

        response: Dict[str, Dict] = self.client.execute_query(query, variables)
        return {
            "orgId": org_id,
            "projectId": response["createProjectSimple"]["project"]["projectId"],
        }

    def get_org(self, org_id: str) -> Dict:
        """Get organization."""
        query = """
            query getOrgSDK($orgId: UUID!) {
                organization(orgId: $orgId){
                    name
                    orgId
                }
            }
        """
        response: Dict[str, Dict] = self.client.execute_query(query, {"orgId": org_id})
        return response["organization"]

    def get_projects(self, org_id: str) -> List[Dict]:
        """Get all projects in organization."""
        query = """
            query getProjectsSDK($orgId: UUID!) {
                projects(orgId: $orgId) {
                    orgId
                    name
                    projectId
                    status
                    desc
                    projectUrl
                }
            }
        """
        response: Dict[str, List[Dict]] = self.client.execute_query(
            query, {"orgId": org_id}
        )
        return response["projects"]

    def get_taxonomies(self, org_id: str) -> List[Dict]:
        """Get a list of taxonomies."""
        query = f"""
            query getTaxonomiesSDK($orgId: UUID!) {{
                taxonomies(orgId: $orgId) {{
                    {TAXONOMY_SHARD}
                }}
            }}
        """
        response: Dict[str, List[Dict]] = self.client.execute_query(
            query, {"orgId": org_id}
        )
        return response["taxonomies"]

    def delete_project(self, org_id: str, project_id: str) -> None:
        """Delete Project."""
        query = """
            mutation removeProjectSDK($orgId: UUID!, $projectId: UUID!) {
                removeProject(orgId: $orgId, projectId: $projectId) {
                    ok
                }
            }
        """
        self.client.execute_query(query, {"orgId": org_id, "projectId": project_id})

    def get_labeling_information(
        self,
        org_id: str,
        start_date: datetime,
        end_date: datetime,
        first: int,
        cursor: Optional[str] = None,
    ) -> Tuple[List[Dict], Optional[str]]:
        """Get org labeling information."""
        query_string = """
        query firstLabelingTimeSDK(
            $orgId: UUID!
            $startDate: DateTime!
            $endDate: DateTime!
            $first: Int
            $after: String
        ) {
            firstLabelingTime(
                orgId: $orgId
                startDate: $startDate
                endDate: $endDate
                first: $first
                after: $after
            ) {
                entries {
                    project {
                        projectId
                    }
                    taskId
                    user {
                        email
                    }
                    timeSpent
                    date
                }
                cursor
            }
        }
        """
        query_variables = {
            "orgId": org_id,
            "startDate": start_date.isoformat(),
            "endDate": end_date.isoformat(),
            "first": first,
            "after": cursor,
        }
        result = self.client.execute_query(query_string, query_variables)
        tasks_paged = result.get("firstLabelingTime", {}) or {}
        entries: List[Dict] = tasks_paged.get("entries", []) or []  # type: ignore
        return entries, tasks_paged.get("cursor")

    def create_taxonomy(
        self,
        org_id: str,
        name: str,
        categories: List[Dict],
        attributes: Optional[List[Dict]],
        task_categories: Optional[List[Dict]],
        task_attributes: Optional[List[Dict]],
    ) -> bool:
        """Create taxonomy."""
        query_string = """
        mutation createTaxonomySDK(
            $orgId: UUID!
            $name: String!
            $categories: [CategoryRootInput!]!
            $attributes: [AttributeInput]
            $taskCategories: [CategoryRootInput!]
            $taskAttributes: [AttributeInput!]
        ) {
            createTaxonomy(
                orgId: $orgId
                name: $name
                categories: $categories
                attributes: $attributes
                taskCategories: $taskCategories
                taskAttributes: $taskAttributes
            ) {
                ok
            }
        }
        """
        query_variables = {
            "orgId": org_id,
            "name": name,
            "categories": categories,
            "attributes": attributes,
            "taskCategories": task_categories,
            "taskAttributes": task_attributes,
        }
        result = self.client.execute_query(query_string, query_variables, False)
        return bool(result and result.get("createTaxonomy", {}).get("ok"))

    def get_label_storage(self, org_id: str, project_id: str) -> Tuple[str, str]:
        """Get label storage method for a project."""
        query_string = """
        query getLabelStorageSDK($orgId: UUID!, $projectId: UUID!) {
            getLabelStorage(orgId: $orgId, projectId: $projectId) {
                storageId
                path
            }
        }
        """
        query_variables = {"orgId": org_id, "projectId": project_id}
        result = self.client.execute_query(query_string, query_variables)
        return (
            result["getLabelStorage"]["storageId"],
            result["getLabelStorage"]["path"],
        )

    def set_label_storage(
        self, org_id: str, project_id: str, storage_id: str, path: str
    ) -> bool:
        """Set label storage method for a project."""
        query_string = """
        mutation updateLabelStorageSDK(
            $orgId: UUID!
            $projectId: UUID!
            $storageId: UUID!
            $path: String!
        ) {
            updateLabelStorage(
                orgId: $orgId
                projectId: $projectId
                storageId: $storageId
                path: $path
            ) {
                ok
            }
        }
        """
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "storageId": storage_id,
            "path": path,
        }
        result = self.client.execute_query(query_string, query_variables, False)
        return bool(result and result.get("updateLabelStorage", {}).get("ok"))



"""Abstract interface to upload."""

from typing import List, Dict, Optional, Any
from abc import ABC, abstractmethod

from pathlib import Path

import aiohttp


class UploadControllerInterface(ABC):
    """Abstract interface to define methods for Upload."""

    @abstractmethod
    def create_datapoint(
        self,
        org_id: str,
        project_id: str,
        storage_id: str,
        name: str,
        items: List[str],
        labels: Optional[List[Dict]],
    ) -> str:
        """
        Create a datapoint and returns its taskId.

        Name must be unique in the project.
        """

    @abstractmethod
    async def create_datapoint_async(
        self,
        aio_client: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        storage_id: str,
        name: str,
        items: List[str],
        labels_data: Optional[str],
        labels_map: Optional[List[Dict]] = None,
        is_ground_truth: bool = False,
    ) -> Dict:
        """
        Create a datapoint and returns its taskId.

        Name must be unique in the project.
        """

    @abstractmethod
    async def item_upload_async(
        self,
        aio_client: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        storage_id: str,
        data_type: str,
        task_type: str,
        file_path: str,
        file_name: str,
        file_type: str,
        is_ground_truth: bool = False,
    ) -> Dict:
        """Upload an item."""

    @abstractmethod
    def upload_image(
        self,
        org_id: str,
        project_id: str,
        file_path: Path,
    ) -> str:
        """Upload a local image and add labels."""

    @abstractmethod
    async def upload_image_async(
        self,
        org_id: str,
        project_id: str,
        file_path: Path,
    ) -> str:
        """Upload a local image and add labels."""

    @abstractmethod
    def items_upload_presign(
        self, org_id: str, project_id: str, files: List[str], file_type: List[str]
    ) -> List[Dict[Any, Any]]:
        """Get a presigned url for uploading items."""

    @abstractmethod
    def delete_tasks(self, org_id: str, project_id: str, task_ids: List[str]) -> bool:
        """Delete tasks in a project."""
